# ============================================================
# Система: Единая библиотека, Центр ИИ НИУ ВШЭ
# Модуль: MLComponent
# Авторы: Полежаев В.А., Хританков А.С.
# Дата создания: 2024 г.
# ============================================================
import logging
import os

from unip.mlcmp.exceptions import UnipMlcmpException

enable_s3_path_correction_var = os.getenv('ENABLE_S3_PATH_CORRECTION', 'True')
ENABLE_S3_PATH_CORRECTION = False if enable_s3_path_correction_var == 'False' else True

enable_s3_url_correction_var = os.getenv('ENABLE_S3_URL_CORRECTION', 'True')
ENABLE_S3_URL_CORRECTION = False if enable_s3_url_correction_var == 'False' else True

MODEL_S3_MODE = os.getenv('MODEL_S3_MODE', 'COPY')  # or 'MOUNT'

# model s3 configuration required for model copy mode
MODEL_S3_BUCKET_NAME = os.getenv('MODEL_S3_BUCKET_NAME')
MODEL_S3_AWS_ACCESS_KEY_ID = os.getenv('MODEL_S3_AWS_ACCESS_KEY_ID')
MODEL_S3_AWS_SECRET_ACCESS_KEY = os.getenv('MODEL_S3_AWS_SECRET_ACCESS_KEY')
MODEL_S3_HOST_PORT = os.getenv('MODEL_S3_HOST_PORT')

# copy model s3 configuration
MODEL_S3_COPY_BUCKET_PATH = os.getenv('MODEL_S3_COPY_BUCKET_PATH')
MODEL_S3_COPY_LOCAL_PATH = os.getenv('MODEL_S3_COPY_LOCAL_PATH')

# no s3 configuration for model mount mode

# model usage configuration
INFERENCE_MODEL_PATH = os.getenv('INFERENCE_MODEL_PATH')
INFERENCE_MODEL_KIND = os.getenv('INFERENCE_MODEL_KIND')
INFERENCE_FUNCTION = os.getenv('INFERENCE_FUNCTION')


FILES_S3_MODE = os.getenv('FILES_S3_MODE', 'COPY')

# all modes files s3 configuration
FILES_S3_BOX_NAME = os.getenv('FILES_S3_BOX_NAME')

# copy files s3 configuration
FILES_S3_COPY_INFERENCE_FILES_DIR = os.getenv('FILES_S3_COPY_INFERENCE_FILES_DIR')

UNIP_MLCMP_APP_NAME = os.getenv('UNIP_MLCMP_APP_NAME')
UNIP_MLCMP_PATH_PREFIX = os.getenv('UNIP_MLCMP_PATH_PREFIX')
UNIP_FILES_API = os.getenv('UNIP_FILES_API')

# license file s3 configuration

# в будущем по аналогии с MODEL_S3_*
# LICENSE_S3_MODE = os.getenv('MODEL_S3_MODE', 'COPY')  # or 'MOUNT'
LICENSE_LOCAL_PATH = os.getenv('LICENSE_LOCAL_PATH')

class InvalidConfiguration(UnipMlcmpException):
    pass

def _validate_files_s3_copy_mode_config():
    if FILES_S3_MODE == 'COPY':
        if FILES_S3_COPY_INFERENCE_FILES_DIR is None:
            raise InvalidConfiguration('FILES_S3_COPY_INFERENCE_FILES_DIR not set, but required in "COPY" mode')
        if FILES_S3_BOX_NAME is None:
            raise InvalidConfiguration('FILES_S3_BOX_NAME not set, but required in "COPY" mode')

def _validate_files_s3_config():
    if FILES_S3_MODE is None:
        raise InvalidConfiguration('FILES_S3_MODE not set')
    if FILES_S3_MODE not in {'COPY'}:
        raise InvalidConfiguration(f'FILES_S3_MODE must be "COPY", given {FILES_S3_MODE}')

    _validate_files_s3_copy_mode_config()


def _validate_model_s3_copy_mode_config():
    if MODEL_S3_MODE != 'COPY':
        return

    global MODEL_S3_HOST_PORT
    global MODEL_S3_COPY_BUCKET_PATH
    logger = logging.getLogger()

    if not all((MODEL_S3_BUCKET_NAME, MODEL_S3_AWS_ACCESS_KEY_ID,
                MODEL_S3_AWS_SECRET_ACCESS_KEY, MODEL_S3_HOST_PORT,
                MODEL_S3_COPY_LOCAL_PATH)):
        raise InvalidConfiguration(f'any of MODEL_S3_BUCKET_NAME, MODEL_S3_AWS_ACCESS_KEY_ID, '
                                   'MODEL_S3_AWS_SECRET_ACCESS_KEY, MODEL_S3_HOST_PORT,'
                                   'MODEL_S3_COPY_LOCAL_PATH is None or empty')

    if MODEL_S3_HOST_PORT.lower().startswith('http://') or MODEL_S3_HOST_PORT.lower().startswith('https://'):
        logger.warning(f'MODEL_S3_HOST_PORT starts with leading "http://" or "https://",'
                       f'but no protocol needed')
        if ENABLE_S3_URL_CORRECTION:
            logger.warning('URL correction is enabled. Leading protocol is removed from MODEL_S3_HOST_PORT.'
                           'Set env ENABLE_S3_URL_CORRECTION="False" to disable URL correction')
            MODEL_S3_HOST_PORT = MODEL_S3_HOST_PORT.replace('https://', '').replace('HTTPS://', '') \
                .replace('http://', '').replace('HTTP://', '')

    if MODEL_S3_COPY_BUCKET_PATH is not None:
        if MODEL_S3_COPY_BUCKET_PATH.startswith('/') or MODEL_S3_COPY_BUCKET_PATH.startswith('\\'):
            logger.warning(f'MODEL_S3_COPY_BUCKET_PATH has leading slash or backward slash,'
                           f'but it means empty bucket "folder" at the bucket root')
            if ENABLE_S3_PATH_CORRECTION:
                logger.warning(
                    'Path correction is enabled. Leading slash is removed from MODEL_S3_COPY_BUCKET_PATH.'
                    'Set env ENABLE_S3_PATH_CORRECTION="False" to disable path correction')
                MODEL_S3_COPY_BUCKET_PATH = MODEL_S3_COPY_BUCKET_PATH[1:]
    else:
        MODEL_S3_COPY_BUCKET_PATH = ''
    if MODEL_S3_COPY_BUCKET_PATH == '':
        logger.warning('MODEL_S3_COPY_BUCKET_PATH is empty, all bucket data from bucket root will be copied')


def _validate_model_s3_config():
    if MODEL_S3_MODE is None:
        raise InvalidConfiguration('MODEL_S3_MODE not set')

    if MODEL_S3_MODE not in {'MOUNT', 'COPY'}:
        raise InvalidConfiguration(f'MODEL_S3_MODE must be "MOUNT" or "COPY", given {MODEL_S3_MODE}')

    if not all((INFERENCE_FUNCTION, INFERENCE_MODEL_PATH)):
        raise InvalidConfiguration(f'any of INFERENCE_FUNCTION, INFERENCE_MODEL_PATH is None or empty')

    _validate_model_s3_copy_mode_config()


def _validate_general_config():
    if UNIP_MLCMP_APP_NAME is None:
        raise InvalidConfiguration('UNIP_MLCMP_APP_NAME not set, but required')
    if UNIP_FILES_API is None:
        raise InvalidConfiguration('UNIP_FILES_API not set, but required')


def _validate_config():
    _validate_general_config()
    _validate_files_s3_config()
    _validate_model_s3_config()


_validate_config()
