# ============================================================
# Система: Единая библиотека, Центр ИИ НИУ ВШЭ
# Модуль: MLComponent
# Авторы: Хританков А.С.
# Дата создания: 2024 г.
# ============================================================
from typing import Any


# noinspection PyUnresolvedReferences
def run_inference_v2(parameters: list,
                     inputs: list, 
                     output_fields: set, 
                     model_key: str,
                     model: Any = None) -> (list, Any):
    """
    Интегрирует модуль с предсказательной моделью в сервис запуска моделей через API.

    Модуль предоставляет API, который имеет принимать именованные входные данные,
    и возвращает также именованные данные.

    Получает на вход несколько параметров вызова :param parameters: как мультимножество, например, так
    ``[['id', 42], ['threshold', 0.001], ['alpha', 1], ['alpha', 2]]``. Все значения - простые типы - строки, числа или логические true/false.


    Входные наборы данных :param inputs: идут с API запросом пользователя, каждый входной набор имеет вид
    ``{"name":"X", "datetype":"FP32", "shape":[3, 1], "data": [1 2 3]}``.

    Примеры типов данных:

    * ``FP32`` - число с плавающей точкой, float32
    * ``CSVFILE`` - путь до CSV файла
    * ``INT32`` - целое число, int32


    Вернуть нужно только поля, переданные в :param output_fields:.
    Примеры выходных наборов данных, которые могут быть возвращены:

    * ``predict`` - вернуть предсказания на входных данных
    * ``predict_proba`` - вернуть оценки вероятностей правильного решения на входных данных
    * ``performance`` - вернуть Python словарик с оценками качества предсказаний от модели
    * ``shapley_values`` - вернуть значения коэффициентов Шепли о влиянии признаков

    Вариант с передачей модели может быть изменен.

    Для расчетов использовать модель, загружаемую из :param model_key:.
    Способ загрузки специфичен для модуля.
    Модель можно вернуть в результате, тогда она может быть передана со следующим запросом.

    Реализация процедуры должна:

    * быть идемпотентой, при повторном вызове вернуть тот же результат
    * не сохранять состояние, зависящее от содержания запросов и входных данных где-то локально
    * не предполагать определенного порядка во входных данных или порядка вызова

    """
    outputs = []
    # желательно использовать ранее загруженную модель, если допустимо
    if not model or model.model_key != model_key:
        model = xxx.load(model_key, **dict(parameters))

    X = {}

    for input in inputs:
        if input.datatype == "text/csv":
            X[input.name] = csv.load(input.data)
        if input.datatype == "FP32":
            X[input.name] = numpy.asarray(input.data).reshape(input.shape)

    if 'predict' in output_fields:
        y_pred = model.predict(**X, **dict(parameters))
        outputs.append(dict(name="predict", 
                            datatype="INT32",
                            shape=y_pred.shape,
                            data=y_pred.astype(int).flatten().tolist()))
        
    if 'predict_proba' in output_fields:
        y_proba = model.predict_proba(**X, **parameters)
        outputs.append(dict(name="predict_proba", 
                            datatype="FP32",
                            shape=y_pred.shape,
                            data=y_pred.astype(float).flatten().tolist()))
        
    if 'performance' in output_fields:
        # e.g. prediction quality
        # datatype = dict
        ...

    if 'shapley_figure' in output_fields:
        shap = yyy.calcShap(model)
        plt.saveFig(shap.toFig(output_path))

        outputs.append(dict(name="shapley_figure", 
                            datatype="image/png",
                            shape=output_size,
                            data=output_path))

    return outputs, model
    
