# ============================================================
# Система: Единая библиотека, Центр ИИ НИУ ВШЭ
# Модуль: MLComponent
# Авторы: Полежаев В.А., Хританков А.С.
# Дата создания: 2024 г.
# ============================================================
import logging
import os
import re
import sys

import json
from typing import Literal, Optional, List

from pydantic import BaseModel
from docstring_parser import DocstringParam, parse


class ApiVar(BaseModel):
    kind: Literal["input", "output"]
    name: str
    description: Optional[str] = None
    datatypes: List[Literal["FILE", "FP32", "FP64", "INT32", "str", "dict"]]
    content_types: Optional[List[str]] = None
    required: Optional[bool] = None


class ApiParam(BaseModel):
    name: str
    description: Optional[str] = None
    param_types: List[str]


def _get_shape(api_var: ApiVar):
    var_char = 'входной' if api_var.kind == 'input' else 'выходной'

    description = f'Размерность значения {var_char} переменной.\n' \
                  'Может быть массивом целых чисел или целым числом.\n' \
                  f'Применимо к любым типам данных {var_char}переменной.\n'
    return {
        'description': description,
        'oneOf': [
            {
                'type': 'array',
                'items': {
                    'type': 'integer'
                },
                'minItems': 1,
                'maxItems': 32
            },
            {
                'type': 'integer'
            }
        ]
    }


def _get_output_field_data(api_var: ApiVar):
    if 'FILE' in api_var.datatypes:
        description = 'Локация, куда следует сохранить выходной файл при datatype == \'FILE\'.\n\n' \
                      'Если значение data соответствует файлу, то значение может быть\n' \
                      'сохранено в указанный файл.\n\n' \
                      'Если значение data соответствует файловой группе, то значение может быть \n' \
                      'сохранено в созданный временный файл в заданной файловой группе.\n\n' \
                      'Если значение data не задано, то значение может быть \n' \
                      'сохранено в созданный временный файл в созданной временной файловой группе.\n\n' \
                      'Выражение \'может быть\' означает, что описанный сценарий будет выполнен, \n' \
                      'если инференс функция не подменит переданное ей в атрибуте data значение.\n'
        return {
            'type': 'string',
            'description': description
        }
    return None


def _get_var_data(api_var: ApiVar):
    var_char = 'входной' if api_var.kind == 'input' else 'выходной'
    description = f'Значение {var_char} переменной.\n'
    if 'FILE' in api_var.datatypes:
        description += 'При datatype == \'FILE\' содержит строку с локацией файла.\n'
    if any(dt != 'FILE' for dt in api_var.datatypes):
        description += 'При datatype != \'FILE\' содержит объект как сериализованную строку.\n'
    return {
        'description': description
    }


def _get_name(api_var: ApiVar):
    var_char = 'входной' if api_var.kind == 'input' else 'выходной'
    return {
        'const': api_var.name,
        'description': f'Имя {var_char} переменной.'
    }


def _get_output_field_datatype(api_var: ApiVar):
    return {
        'type': 'string',
        'description': f'Ожидаемый тип выходной переменной.',
        'enum': api_var.datatypes
    }


def _get_var_datatype(api_var: ApiVar):
    var_char = 'входной' if api_var.kind == 'input' else 'выходной'
    return {
        'type': 'string',
        'description': f'Тип {var_char} переменной.',
        'enum': api_var.datatypes
    }


def _get_output_field_content_type(api_var: ApiVar):
    if api_var.content_types is None:
        return None
    spec = {
        'type': 'string',
        'description': 'Ожидаемый тип содержимого файла, \n'
                       'при ожидаемом типе выходной переменной datatype == \'FILE\'.\n',
    }
    if len(api_var.content_types) > 0:
        spec['enum'] = api_var.content_types
    return spec


def _get_var_content_type(api_var: ApiVar):
    if api_var.content_types is None:
        return None
    var_char = 'входной' if api_var.kind == 'input' else 'выходной'
    spec = {
        'type': 'string',
        'description': 'Тип содержимого файла, \n'
                       f'при типе {var_char} переменной datatype == \'FILE\'.\n'
    }
    if len(api_var.content_types) > 0:
        spec['enum'] = api_var.content_types
    return spec


def _get_var_base(description: str = None):
    openapi_spec = {
        'type': 'object',
        'properties': {},
        'required': [
            'name',
            'datatype'
        ]
    }
    if description:
        openapi_spec['description'] = description

    return openapi_spec


def _construct_output_field(api_var: ApiVar):
    openapi_spec = _get_var_base(api_var.description)
    openapi_spec['properties']['name'] = _get_name(api_var)
    openapi_spec['properties']['datatype'] = _get_output_field_datatype(api_var)
    data_spec = _get_output_field_data(api_var)
    if data_spec:
        openapi_spec['properties']['data'] = data_spec
    content_type = _get_output_field_content_type(api_var)
    if content_type:
        openapi_spec['properties']['content_type'] = content_type
    return openapi_spec


def _construct_var(api_var: ApiVar):
    openapi_spec = _get_var_base(api_var.description)
    openapi_spec['properties']['name'] = _get_name(api_var)
    openapi_spec['properties']['datatype'] = _get_var_datatype(api_var)
    openapi_spec['properties']['data'] = _get_var_data(api_var)
    openapi_spec['required'].append('data')
    shape = _get_shape(api_var)
    openapi_spec['properties']['shape'] = shape
    openapi_spec['required'].append('shape')
    if api_var.content_types is not None:
        content_type = _get_var_content_type(api_var)
        if content_type:
            openapi_spec['properties']['content_type'] = content_type
    return openapi_spec


def _construct_model_param(api_param: ApiParam):
    openapi_spec = {
        'type': 'array',
        'prefixItems': [
            {
                'const': api_param.name
            },
            {
                'type': 'object',
                'additionalProperties': {
                    'type': api_param.param_types
                }
            }
        ],
        'minItems': 2,
        'maxItems': 2
    }
    if api_param.description:
        openapi_spec['description'] = api_param.description

    return openapi_spec


def _parse_docstring_var_name(name_str: str):
    parts = name_str.split('.')
    if len(parts) != 2:
        return None, None
    if parts[0] not in {'inputs', 'outputs', 'model_parameters'}:
        return None, None
    parts[0] = parts[0][:-1]  # remove trailing 's'
    return parts


def _parse_docstring_var_type(type_str: str, supported_datatypes):
    required = True
    if type_str and type_str.endswith('optional'):
        required = False
        type_str = type_str.rstrip('optional').rstrip(' ').rstrip(',')

    if type_str:
        parts = re.split(r'or\s+', type_str)
        parts[0].strip()

        parts = [re.sub(r'[\s"\']*$', '', p) for p in parts]
        parts = [re.sub(r'^[\s"\']*', '', p) for p in parts]
        parts = set(parts)
        datatypes = parts & set(supported_datatypes)
        content_types = parts - datatypes
        if len(content_types) == 0:
            content_types = None
    else:
        datatypes = set(supported_datatypes)
        content_types = set()

    if content_types is not None and ('FILE' not in datatypes):
        datatypes.add('FILE')
    return datatypes, content_types, required


def _parse_docstring_param_type(type_str: str, supported_param_types):
    if not type_str:
        return ['string']

    if type_str.endswith('optional'):
        type_str = type_str.rstrip('optional').rstrip(' ').rstrip(',')

    parts = re.split(r'or\s+', type_str)
    parts[0].strip()

    parts = [re.sub(r'[\s"\']*$', '', p) for p in parts]
    parts = [re.sub(r'^[\s"\']*', '', p) for p in parts]
    parts = set(parts)
    param_types = parts & set(supported_param_types)
    if not param_types:
        return ['string']
    return param_types


def _create_var(var_name, var_kind, param: DocstringParam,
                supported_datatypes) -> ApiVar:
    dict_var = {
        'name': var_name,
        'kind': var_kind,
        'description': param.description
    }

    datatypes, content_types, required = _parse_docstring_var_type(param.type_name,
                                                                   supported_datatypes)
    dict_var['datatypes'] = datatypes
    # if content_types is not None:
    dict_var['content_types'] = content_types
    if dict_var['kind'] == 'input':
        dict_var['required'] = required

    api_var = ApiVar(**dict_var)
    return api_var


def _create_param(var_name: str, param: DocstringParam, supported_param_types) -> ApiParam:
    dict_param = {
        'name': var_name,
        'description': param.description,
        'param_types': _parse_docstring_param_type(param.type_name, supported_param_types)
    }

    api_param = ApiParam(**dict_param)
    return api_param


def _create_spec_from_docstring(docstring: str, supported_datatypes, supported_param_types):
    logger = logging.getLogger(__name__)
    logger.info('OpenAPI Spec building from docstring started...')
    docstring_struct = parse(docstring)
    processed_params = set()
    api_vars = []
    api_model_params = []
    for param in docstring_struct.params:
        var_key = param.arg_name
        logger.info(f'Found param {param.arg_name}')
        if var_key in processed_params:
            logger.info(f'{param.arg_name} already processed, skipped')
            continue
        var_kind, var_name = _parse_docstring_var_name(param.arg_name)
        logger.info(f'Param {param.arg_name} parsed; kind: "{var_kind}"; name: "{var_name}"')
        if not var_name:
            continue
        logger.info(f'Processing var "{var_name}"')
        if var_kind == 'input' or var_kind == 'output':
            api_var = _create_var(var_name, var_kind, param, supported_datatypes)
            api_vars.append(api_var)
        if var_kind == 'model_parameter':
            api_param = _create_param(var_name, param, supported_param_types)
            api_model_params.append(api_param)
        processed_params.add(var_key)

    required_input_vars = [spec_var for spec_var in api_vars
                           if spec_var.required and spec_var.kind == 'input']
    required_input_vars_names = [spec_var.name for spec_var in required_input_vars]
    required_input_vars_str = f'Обязательные переменные: {", ".join(required_input_vars_names)}'
    optional_input_vars = [spec_var for spec_var in api_vars
                           if not spec_var.required and spec_var.kind == 'input']
    output_vars = [spec_var for spec_var in api_vars
                   if spec_var.kind == 'output']

    inputs = [_construct_var(v) for v in required_input_vars]
    inputs_min_items = len(inputs)
    inputs += [_construct_var(v) for v in optional_input_vars]
    outputs = [_construct_var(v) for v in output_vars]
    output_fields = [_construct_output_field(v) for v in output_vars]
    model_parameters = [_construct_model_param(p) for p in api_model_params]

    inputs_spec, output_fields_spec, outputs_spec, model_parameters_spec = (None,) * 4

    if outputs:
        outputs_spec = {
            'description': 'Выходные переменные, возвращаются как результат из инференс функции.\n',
            'type': 'array',
            'items': {'oneOf': outputs}
        }

    if output_fields:
        output_fields_spec = {
            'description': 'Список выходных переменных, которые инференс функция должна вернуть.',
            'type': 'array',
            'items': {'oneOf': output_fields}
        }

    if inputs:
        inputs_spec = {
            'type': 'array',
            'description': f'Входные переменные, передаются как аргумент в инференс функцию.'
                           f'\n{required_input_vars_str}',
            'items': {'oneOf': inputs},
            'minItems': inputs_min_items
        }

    if model_parameters:
        # это нужно для параметров модели, которые не специфицированы явно в схеме
        any_param = {
            'type': 'array',
            'description': 'Произвольный параметр модели',
            'prefixItems': [
                {'type': 'string'},
                {}
            ],
            'minItems': 2,
            'maxItems': 2
        }
        model_parameters_spec = {
            'type': 'array',
            'description': 'Параметры, передаются как аргумент в инференс функцию.\n'
                           'Каждый параметр представляется кортежем из двух элементов (имя, значение).\n',
            'items': {'oneOf': model_parameters + [any_param]}
        }

    logger.info('OpenAPI Spec building from docstring finished')
    logger.info(f'Input vars found: {len(inputs)}; output vars found: {len(outputs)}; '
                f'model parameters found: {len(model_parameters)}')
    return inputs_spec, output_fields_spec, outputs_spec, model_parameters_spec


def create_openapi_spec(inference_function, datatypes, param_types) -> dict:
    base_dir = os.path.dirname(os.path.realpath(__file__))
    json_path = os.path.join(base_dir, 'openapi-base.json')

    with open(json_path) as f:
        spec_dict = json.load(f)

    inputs_spec, output_fields_spec, outputs_spec, params_spec = \
        _create_spec_from_docstring(inference_function.__doc__, datatypes, param_types)

    schemas = spec_dict['components']['schemas']
    req_properties = schemas['PredictionRequest']['properties']
    if inputs_spec:
        req_properties['inputs'] = inputs_spec
    if output_fields_spec:
        req_properties['output_fields'] = output_fields_spec
    if params_spec:
        req_properties['model_parameters'] = params_spec
    if outputs_spec:
        resp_properties = schemas['PredictionResponse']['properties']
        resp_properties['outputs'] = outputs_spec

    return spec_dict


if __name__ == '__main__':

    from typing import Any, Literal, Optional


    def example_inference_func(parameters: list,
                               inputs: list,
                               output_fields: set,
                               model_key: str,
                               model: Any = None) -> (list, Any):
        """
        :param inputs.p: Параметр прогнозирования;
            Если вход опциональный, то строка типа завершается ', optional',
            это соответствует синтаксису sphinx rst
        :type inputs.p: FP32, optional

        :param inputs.X: Матрица экземпляров для классификации;
            Если типов несколько, они перечисляются через 'or',
            это соответствует синтаксису sphinx rst
        :type inputs.X: FP64 or FP32 or FILE or text/csv

        :param inputs.abc: Входной параметр;
            Тип может быть не указан, тогда считается,
            что поддерживаются все допустимые datatype и любые content-type
        :type inputs.abc:

        :param outputs.predict: Результат прогнозирования
        :type outputs.predict: INT32 or 'FILE' or text/csv or application/json or BED

        :param outputs.scores: Скоры прогнозирования;
            Для этого выхода datatype 'FILE' будет добавлен, потому что указан content-type 'text/csv'
        :type outputs.scores: FP32 or FP64 or text/csv

        :param outputs.obj: Только объектный выход;
            Для этого выхода тип только объектный
        :type outputs.obj: INT32 or str, optional

        :param model_parameters.mode: Режим модели;
            Допустимые типы параметров модели - примитивные типы OpenAPI Spec
        :type model_parameters.mode: number or integer or float

        :param model_parameters.param1: Параметр работы модели
        :type model_parameters.param1: string, optional
        """
        pass


    def import_inference_func(func_name):
        index = func_name.rindex('.')
        from importlib import import_module
        func_module = import_module(func_name[:index])
        func_name = func_name[index + 1:]
        func = getattr(func_module, func_name)
        return func


    if len(sys.argv) != 2 and len(sys.argv) != 3:
        raise SyntaxError('At least output path must be passed; inference function full name can be passed to')
    if len(sys.argv) == 3:
        inference_func = import_inference_func(sys.argv[2])
    else:
        inference_func = example_inference_func

    spec = create_openapi_spec(inference_func,
                               ['FILE', 'FP32', 'FP64', 'dict', 'str', 'INT32'],
                               ['boolean', 'integer', 'number', 'string'])

    with open(sys.argv[1], 'w', encoding='utf-8') as f:
        json.dump(spec, f, ensure_ascii=False, indent=2)
