# ============================================================
# Система: Единая библиотека, Центр ИИ НИУ ВШЭ
# Модуль: MLComponent
# Авторы: Полежаев В.А., Хританков А.С.
# Дата создания: 2024 г.
# ============================================================
from typing import Optional, List, Union, Tuple, Any, Literal

from pydantic import BaseModel
from pydantic.version import VERSION as pydantic_version

DATATYPES = Literal["FILE", "FP32", "FP64", "INT32", "dict", "str"]
CONTENT_TYPES = Literal["text/csv", "application/json"]
PARAM_TYPES = ["integer", "number", "string", "boolean"]


class Input(BaseModel):
    name: str
    data: Any
    datatype: DATATYPES
    content_type: Optional[str] = None
    shape: Union[int, List[int]]


class OutputField(BaseModel):
    name: str
    data: Optional[str] = None
    datatype: DATATYPES
    content_type: Optional[str] = None


if pydantic_version >= '2':
    from pydantic import ConfigDict

    class PredictRequest(BaseModel):
        model_parameters: Optional[List[Tuple[Any, Any]]] = None
        inputs: List[Input]
        output_fields: List[OutputField]
        model_config = ConfigDict(
            protected_namespaces=()
        )
else:
    class PredictRequest(BaseModel):
        model_parameters: Optional[List[Tuple[Any, Any]]] = None
        inputs: List[Input]
        output_fields: List[OutputField]


class Output(BaseModel):
    name: str
    data: Any  # presigned_get_url for file
    datatype: DATATYPES
    content_type: Optional[str] = None
    shape: Union[int, List[int]]


class PredictResponse(BaseModel):
    outputs: List[Output]


if pydantic_version >= '2':
    from pydantic import ConfigDict

    class ModelVersionResponse(BaseModel):
        openapi_spec: Optional[dict]
        license: Optional[str]
        model_version: str = '1'
        model_config = ConfigDict(
            protected_namespaces=()
        )
else:
    class ModelVersionResponse(BaseModel):
        openapi_spec: Optional[dict]
        license: Optional[str]
        model_version: str = '1'


class UnipMlcmpException(Exception):
    pass


class InputValdationException(UnipMlcmpException):

    def __init__(self, *args, validation_errors=None):
        super().__init__(*args)
        self.validation_errors = validation_errors


class UnauthenticatedException(UnipMlcmpException):
    pass


class InferenceException(UnipMlcmpException):
    pass


class InputFileNotFoundException(UnipMlcmpException):
    pass


class ResourceNotFoundException(UnipMlcmpException):
    pass
